% This script takes in raw MRI data and creates a binary mask of the skull
% underneath

%File names can be hardcoded in or taken as user input
clearvars
close all
%%%%%%%%%%%%%%%%%%%%%%%%%% SEIZURE WARNING %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Parts of this code are animated and have rapidly flashing black and white
% images through it. Make sure to comment these parts out before running
pauseDur = 0.001; % INCREASE THIS IF YOU ARE A SEIZURE RISK
%%%%%%%%%%%%%%%%%%%%%%%%%% SEIZURE WARNING %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
tic              

%Get file name, can be changed here where name is hard coded 

%name = ('FileName.dcm');
disp('What is the name of the MRI file?')
name = input('','s');
MPRAGE = squeeze(dicomread(name));
MP_info = dicominfo(name);

frame = 41;
%% Adjust threshold for image 
disp('Do you know the threshold for skull extraction (y)es or (n)o?')
threshQ = input('','s');
if threshQ ~= 'y'
    threshold = 200;
    disp(['threshold: ',num2str(threshold)]);
    disp('')
    disp('Calibrate binary threshold using (f)orward / (b)ackward and')
    disp('(h)igher / (l)ower. Enter (y) when finished.')
    ui = 'null';
    figure;
    while ui ~= 'y'
        mask = MPRAGE > threshold;
        pcolor(flipud(mask(:,:,frame))), shading flat, colormap(bone);title(num2str(frame))
        ui = input(' ','s');
        if ui == 'f'
            frame = frame+1;
        elseif ui == 'b'
            frame = frame-1;
        elseif ui == 'h'
            threshold = threshold +5;
        elseif ui == 'l'
            threshold = threshold -5;
        end
        disp(['threshold: ',num2str(threshold)]);
    end
else
    threshold = input('Enter threshold for skull extraction:  ');
    figure;
    mask = MPRAGE > threshold;
    pcolor(flipud(mask(:,:,frame))), shading flat, colormap(bone);title(num2str(frame))
    disp('Are you sure this is the threshold you want to use? (y)es or (n)o')
    ui2 = input('','s');
    if ui2 ~= 'y'
        disp(['threshold: ',num2str(threshold)]);
        disp('')
        disp('Calibrate binary threshold using (f)orward / (b)ackward and')
        disp('(h)igher / (l)ower. Enter (y) when finished.')
        ui = 'null';
        figure;
        while ui ~= 'y'
            mask = MPRAGE > threshold;
            pcolor(flipud(mask(:,:,frame))), shading flat, colormap(bone);title(num2str(frame))
            ui = input(' ','s');
            if ui == 'f'
                frame = frame+1;
            elseif ui == 'b'
                frame = frame-1;
            elseif ui == 'h'
                threshold = threshold +5;
            elseif ui == 'l'
                threshold = threshold -5;
            end
            disp(['threshold: ',num2str(threshold)]);
        end
    end
end

%% Fill locations for each frame
% This section fills in the non monkey area. this does a rough job of
% eliminating the muscluature around the the skull.

% There are certain areas where the "bone" areas bleed into the background,
% so the script runs through two passes:

%the first pass is through the transverse axis
transLoc = {[3 3], [3 size(mask,3)-3], [size(mask,1)-3 3], [size(mask,1)-3 size(mask,3)-3]}; % transLoc specifies which coordinates to originate the fill from. They are the upper corners.
transBim = mask; % Save current mask to a temporary variable

% figure % Quick visualization showing the filling of extra space
% Use this to adjust your threshold values or your infil origin points
% Once this has been calibrated take out pcolor, drawnow, and pause to make
% code run faster
for i = 1:size(transBim,1)
    for j = 1:length(transLoc)
        transBim(i,:,:) = imfill(squeeze(transBim(i,:,:)),transLoc{j}, 8);
    end
end
%% The second pass is through the coronal axis
coroLoc = {[3 3], [3 size(mask,1)-3], [size(mask,1)-3 3], [size(mask,1)-3 size(mask,1)-3]};
coroBim = mask;

% figure
for i = 1:size(coroBim,3)
    for j = 1:length(coroLoc)
        coroBim(:,:,i) = imfill(squeeze(coroBim(:,:,i)),coroLoc{j}, 8);
    end
end

%% Invert Mask, remove noise
% This section inverts the mask to make the bone TRUE this is important for
% isosurface which maps the mask to 3D
transRim = ~transBim;
coroRim = ~coroBim;

% This section removes all isolated pixel objects less than 1000 pixels
% IN VOLUME. This retains only the major structures, like the skull
transObj = bwconncomp(transRim);
coroObj = bwconncomp(coroRim);

transPixels = cellfun(@numel, transObj.PixelIdxList);
coroPixels = cellfun(@numel, coroObj.PixelIdxList);

transIdx = transPixels > 5000;
coroIdx = coroPixels > 5000;

for i = 1:transObj.NumObjects
    if transIdx(i) == 0
        transRim(transObj.PixelIdxList{i}) = 0;
    end
end
       
for i = 1:coroObj.NumObjects
    if coroIdx(i) == 0
        coroRim(coroObj.PixelIdxList{i}) = 0;
    end
end

%% Combine transverse and coronal passes
% Since the masks have now been over-cleaned, this sections stitches both
% passes back together, retaining areas of overlap. This mitigates missing
% sections. If you are still missing more sections than desired, consider
% two options:
% 1) Adjust threshold. This will create less inflow of the fill function
% 2) Add a third saggital pass

cleanIm = coroRim + transRim;
cleanIm(cleanIm == 2) = 1;
pcolor(flipud(cleanIm(:,:,frame))), shading flat, colormap(bone);

%% Perform Brain Extraction
[cleanImB] = brain_extraction(cleanIm,MPRAGE);

%% Scale Image

%voxel size in mm
xsize = MP_info.PerFrameFunctionalGroupsSequence.Item_1.PixelMeasuresSequence.Item_1.PixelSpacing(1);
ysize = MP_info.PerFrameFunctionalGroupsSequence.Item_1.PixelMeasuresSequence.Item_1.PixelSpacing(2);
zsize = MP_info.PerFrameFunctionalGroupsSequence.Item_1.PixelMeasuresSequence.Item_1.SpacingBetweenSlices;

if xsize ~= ysize, error 'this code assumes the x and y voxel lengths are equal',end
scaler = zsize/xsize;

%full scaled image
cleanIm2 = cleanIm;
for n=1:ceil(scaler)-1
    cleanIm2 = cat(3,cleanIm2,cleanIm.*0);%expand the matrix so that there is room to stretch the image
end
[X,Y,Z] = ndgrid(0:size(cleanIm2,1)-1, 0:size(cleanIm2,2)-1, (0:(size(cleanIm2,3)-1)));
scaledNoCrop = griddedInterpolant(X,Y,Z.*scaler,cleanIm2);

%scaled brain

cleanIm2B = cleanImB;
for n=1:ceil(scaler)-1
    cleanIm2B = cat(3,cleanIm2B,cleanImB.*0);%expand the matrix so that there is room to stretch the image
end
[XB,YB,ZB] = ndgrid(0:size(cleanIm2B,1)-1, 0:size(cleanIm2B,2)-1, (0:(size(cleanIm2B,3)-1)));
scaledNoCropB = griddedInterpolant(XB,YB,ZB.*scaler,cleanIm2B);

%% Display
%Skull
figure;
snc = scaledNoCrop(X,Y,Z);
p = patch(isosurface(double(permute(flip(snc,1),[3 2 1]))));%This is for generating images
p.FaceColor = [0.93 0.93 0.93];
p.EdgeColor = 'none';
daspect([1 1 1]);
camlight; 
lighting phong
title('scaled')

%Brain
figure;
sncB = scaledNoCropB(XB,YB,ZB);
p = patch(isosurface(double(permute(flip(sncB,1),[3 2 1]))));%This is for generating images
p.FaceColor = [0 0.4470 0.7410];
p.EdgeColor = 'none';
daspect([1 1 1]);
camlight; 
lighting phong
title('scaledB')

%% Pick location on the brain
disp('Do you already know the coordinates for the center the craniotomy? (y)es or (n)o')
ui = input('','s');
if ui == 'y'
    disp('Please enter the coordinates')
    disp('x = ')
    ui = input('','s');
    y = round(str2num(ui));
    disp('y = ')
    ui = input('','s');
    x = round(str2num(ui));
    disp('z = ')
    ui = input('','s');
    z = round(str2num(ui));
    frame = z;
else
    disp('Select which slice you would like to center the craniotomy on by using the (f)orward or (b)ackward')
    disp('Enter (y) when finished.')
    frame = 41;
    ui = 'null';
    while ui ~= 'y'
        imshow(MPRAGE(:,:,frame),[]), truesize([600 600]);title(["zFrame=",num2str(frame)]);hold on%[y, x] = getpts;x = round(x);y = round(y); title(["x=",num2str(x)," y=",num2str(y),"zFrame=",num2str(frame)]);
        ui = input(' ','s');
        if ui == 'f'
            frame = frame+1;
        elseif ui =='b'
            frame = frame-1;
        end
    end
    selectedCenter=false;
    while selectedCenter ~=1
        disp('Now select on the skull where you would like the craniotomy centered')
        disp('Hit "Enter" after making your selction')
        [y, x] = getpts;
        x = round(x);
        y = round(y);
        z = frame;
        title(["x=",num2str(x)," y=",num2str(y),"zFrame=",num2str(frame)]);
        ui = input('Are you sure this is the location you want? (y)es or (n)o: ','s');
        if ui == 'y'
            selectedCenter = 1;
        end
    end
end
disp('The coordinates for the center of the craniotomy are:')
disp(['x = ',num2str(y)])
disp(['y = ',num2str(x)])
disp(['z = ',num2str(z)])

%% Cut hole from the skull
sphere_center = [x y z]; %x,y,z. EXAMPLE VALUES
crani = cleanIm2;
rad = [];
while(isempty(rad) || rad<0)
    disp(' ')
    rad = input('Please enter the radius of the craniotomy in mm and press enter: '); % Example value: 12.5
end

radius = rad/xsize;%desired radius (mm) / distance between pixels (mm/pix)

ball = crani.*0;
ball((X-sphere_center(1)).^2 + (Y-sphere_center(2)).^2 + ((Z-sphere_center(3)).*scaler).^2 <= radius.^2) = 1;

crani = crani - ball;
crani(crani<0)=0;

ui = input('Would you like a second, outer radius? (y)es or (n)o: ','s');
if ui == 'y'
    rad2 = [];
    while(isempty(rad2) || rad2<rad)
        disp(' ')
        rad2 = input('Please enter the outer radius in mm and press enter: '); % Example value: 25
    end
    
    radius = rad2/xsize;%desired radius (mm) / distance between pixels (mm/pix)

    ball = crani.*0;
    ball((X-sphere_center(1)).^2 + (Y-sphere_center(2)).^2 + ((Z-sphere_center(3)).*scaler).^2 >= radius.^2) = 1;

    crani = crani - ball;
    crani(crani<0)=0;
end

%% Display Visualization
scaledCrani = griddedInterpolant(X,Y,Z.*scaler,crani);
figure;
sci = scaledCrani(X,Y,Z);
p = patch(isosurface(double(permute(flip(sci,1),[3 2 1])))); % This is for generating images
p.FaceColor = [0.93 0.93 0.93];
p.EdgeColor = 'none';
daspect([1 1 1]);
camlight; 
lighting phong

%% Crop scaled craniotomy
cleanIm3 = crani;
cleanIm3 = cleanIm3(1:170,:,:);

scale_bar_selection = input('Would you like a scale bar for skull image? (y)es or (n)o: ','s');
switch scale_bar_selection
    case 'y'
        cleanIm3(1,:,1) = ones(size(cleanIm3(1,:,1))); %adding the scale bar
        disp('Scale bar added')
    otherwise
        % do nothing
end
[X,Y,Z] = ndgrid(0:size(cleanIm3,1)-1, 0:size(cleanIm3,2)-1, 0:(size(cleanIm3,3)-1));
scaledCrop = griddedInterpolant(X,Y,Z.*scaler,cleanIm3);

figure;

p = patch(isosurface(double(permute(flip(scaledCrop(X,Y,Z),1),[3 2 1]))));
p.FaceColor = [0.93 0.93 0.93];
p.EdgeColor = 'none';
daspect([1 1 1]);
camlight; 
lighting phong
title('Scaled and Cropped Skull')
finalArray = scaledCrop(X,Y,Z);
finalArray(finalArray>0.5)=1;
finalArray(finalArray<0.5)=0;

%Crop Brain
cleanIm3B = cleanIm2B;
cleanIm3B = cleanIm3B(1:170,:,:);

scale_bar_selection = input('Would you like a scale bar for brain image? (y)es or (n)o: ','s');
switch scale_bar_selection
    case 'y'
        cleanIm3B(1,:,1) = ones(size(cleanIm3B(1,:,1))); %adding the scale bar
        disp('Scale bar added')
    otherwise
        % do nothing
end
[XB,YB,ZB] = ndgrid(0:size(cleanIm3B,1)-1, 0:size(cleanIm3B,2)-1, 0:(size(cleanIm3B,3)-1));
scaledCropB = griddedInterpolant(XB,YB,ZB.*scaler,cleanIm3B);

figure;

p = patch(isosurface(double(permute(flip(scaledCropB(XB,YB,ZB),1),[3 2 1]))));
p.FaceColor = [0.93 0.93 0.93];
p.EdgeColor = 'none';
daspect([1 1 1]);
camlight; 
lighting phong
title('Scaled and Cropped Brain')
finalArrayB = scaledCropB(XB,YB,ZB);
finalArrayB(finalArrayB>0.5)=1;
finalArrayB(finalArrayB<0.5)=0;
%% save as a convenient file type
disp('Do you want to save these images? (y)es or (n)o')
uifig = input('','s');
if uifig == 'y'
    %fileNameA = ('FileName.stl')
    disp('What would you like the name of the craniotomy b to be?')
    fileNameA = [input('','s') '.stl'];
    make_stl_of_array(fileNameA,flip(scaledCrop.Values),xsize,ysize,zsize);
    disp('Skull file saved!')
    
    %fileNameB = ('BerryBrain.stl')
    disp('What would you like the name of the brain STL to be?')
    fileNameB = [input('','s') '.stl'];
    make_stl_of_array(fileNameB,flip(scaledCropB.Values),xsize,ysize,zsize);
    disp('Brain file saved!')
else
    %Do nothing
end
%% Create Combined image
combinedIm = scaledCrop.Values;
for i = 1:length(scaledCrop.Values(1,1,:)) %go through z
        for j = 1:length(scaledCrop.Values(1,:,1)) %go through x
            for k = 1:length(scaledCrop.Values(:,1,1)) %go through y down from the top
                if scaledCrop.Values(k,j,i) == 1 || scaledCropB.Values(k,j,i) == 1  %if the value in the skull mask is 1
                    combinedIm(k,j,i) = 1;
                else
                   combinedIm(k,j,i) = 0;
                end
            end
        end
end

%%Plot with diff colors
figure;

p = patch(isosurface(double(permute(flip(scaledCrop(X,Y,Z),1),[3 2 1]))));
p.FaceColor = [0.93 0.93 0.93];
p.EdgeColor = 'none';
daspect([1 1 1]);
camlight; 
%lighting phong
title('Brain and Skull with Craniotomy')
finalArray = scaledCrop(X,Y,Z);
finalArray(finalArray>0.5)=1;
finalArray(finalArray<0.5)=0;
hold on
p = patch(isosurface(double(permute(flip(scaledCropB(XB,YB,ZB),1),[3 2 1]))));
p.FaceColor = [0 0.4470 0.7410];
p.EdgeColor = 'none';
daspect([1 1 1]);

%% Prepare to select points for STL with reduced file size
stl_size_reduction = input('Would you like to select points for STL size reduction? (y)es or (n)o: ','s');
if stl_size_reduction=='y'
    
    %collect points
    s=NaN(sum(finalArray(:)),3);
    s_idx=1;
    for i=1:size(finalArray,1)
        for j=1:size(finalArray,2)
            for k=1:size(finalArray,3)
                if finalArray(i,j,k)==1
                    s(s_idx,:)=[i,j,k];
                    s_idx=s_idx+1;
                end
            end
        end
    end
    
    
    %collect points
    input('Add data points to the latest figure. Hit Enter when complete.','s');
    myInfo=getCursorInfo(datacursormode);
    
    % assemble points array
    p=NaN(length(myInfo),3);%p for points, 3 columns for 3 dimensions
    for n=1:length(myInfo)
        p(n,:)=myInfo(n).Position;
    end

    figure
    p=p./scaler; %scale to get the correct dimensions 
    
    %flip one axis to get the correct orientation
    maxpx=round(max(p(:,1))+1);
    minpx=min(p(:,1));
    if minpx<=0
        disp('NOTE this code has not been tested with x values at or less than 0')
    end
    p(:,1)=maxpx-p(:,1);
    
    %wrap points and generate stl
    k=boundary(p);
    patch=trisurf(k,p(:,1),p(:,2),p(:,3),0);
    patch2=struct('vertices',patch.Vertices,'faces',patch.Faces);
    
    %fileNameC = ('FileName.stl')
    disp('What would you like the name of reduced file to be?')
    fileNameC = [input('','s') '.stl'];
    stl_write(fileNameC,patch2);
end
%% Finish out
disp('You are all done. Congrats!')